part of oche;

class _RouteTrie {

  static const String _SEPARATOR = '/';
  _RouteTrieNode _rootNode = new _RouteTrieNode('');

  /**
   * [build] creates a trie of [_RouteTrieNode]s by looping through
   * the [routeMap] and get the routes from the module.
   * For each route, we get the segments its path consists combined
   * with the basePath for the module.
   * For each segment, starting from the [_rootNode] as currentNode,
   * we look for a child node which has the given segment.
   * If found, we take the matchedNode and set it as currentNode.
   * If not found, we create e new node and add it to the childNodes
   * of the currentNode, then set it as currentNode.
   * Then we continue the same steps for each segment.
   * The last currentNode represent the last segment in the route path,
   * and will be given the method, action and parentModule type of the route.
   */
  void build(Map<String, Type> routeMap) {

    routeMap.forEach((String basePath, Type module) {

      List<Route> routes = _getModuleRoutes(module);

      routes.forEach((route) {

       var currentNode = _rootNode;

       var segments = _getPathSegments(route._path, basePath: basePath);

        segments.forEach((segment) {

          var matchedNode = _findChildNode(currentNode, segment);

          if (matchedNode == null) {

            matchedNode = new _RouteTrieNode(segment);

            currentNode.childNodes.add(matchedNode);

          }

          currentNode = matchedNode;

        });

        currentNode
        ..action = route._action
        ..method = route._method
        ..module = route._parentModule.runtimeType;

      });

    });

  }

  /**
   * [findNode] finds a [_RouteTrieNode] in the trie by looping through
   * each segment in the [path] and looks if the segment and
   * the [method] matches any node, starting from the [_rootNode].
   * If none is found it returns null;
   */
  _RouteTrieNode findNode(String path, String method) {

    var currentNode = _rootNode;

    var segments = _getPathSegments(path);

    segments.forEach((segment) {

      if (currentNode != null) {

        var matchedNode = _findChildNode(currentNode, segment, method: method);

        currentNode = matchedNode;

      }

    });

    return currentNode;

  }

  /**
   * [_getModuleRoutes] gets the [Route]s of the [module].
   * If the [Type] of [module]s superclass is not [OcheModule]
   * it will throw a [TypeException].
   */
  List<Route> _getModuleRoutes(Type module) {

   var classMirror = reflectClass(module);

   if (classMirror.superclass.reflectedType != OcheModule) {

     throw new TypeException('Module does not extend OcheModule.');

   }

   var instanceMirror = classMirror.newInstance(new Symbol(''), []);

   var instance = instanceMirror.reflectee as OcheModule;

   return instance._routes;

  }

  /**
   * [_findChildNode] finds a specific child node among the children
   * of [parentNode] by either [segment] or [segment] and [method].
   * If no match is found it returns null;
   */
  _RouteTrieNode _findChildNode(_RouteTrieNode parentNode, String segment, { String method }) {

    _RouteTrieNode childNode;

    if (parentNode.childNodes.length != 0) {

      childNode = parentNode.childNodes.firstWhere((node) {

        if (method != null) {

          return node.segment == segment && node.method == method;

        }

        return node.segment == segment;

      }, orElse: () {

        return null;

      });

    }

    return childNode;

  }

  /**
   * [_getPathSegments] takes a [path] and optional [basePath]
   * and returns a [List] of segments.
   */
  List<String> _getPathSegments(String path, { String basePath: '' }) {

    if (basePath == '/') {

      basePath = '';

    }

    path = basePath + path;

    return new List<String>()
        ..addAll(path.split(_SEPARATOR).skip(1));

  }

}